# -*-python-*-

__package__    = "s60-restobook"
__version__    = "0.55"
__author__     = "Aaron Straup Cope"
__url__        = "http://www.aaronland.info/python/restobook/"
__cvsversion__ = "$Revision: 1.54 $"
__date__       = "$Date: 2006/09/13 04:09:34 $"
__copyright__  = "Copyright (c) 2006 Aaron Straup Cope. Perl Artistic License."

import re
import e32
import anydbm
import appuifw
import base64
import md5
import os.path

if not e32.in_emulator() :
    import contacts

from elementtree.ElementTree import XML
from pdis.xpath import compile
import httplib
import urllib

class Resto :

    def __init__ (self) :

        self.__msg = ""

        self.__network = False
        self.__offline = False

        self.__tags   = None
        self.__search = []
        
        self.__cfgpath = u"c:\\restobook.db"     
        self.lock  = e32.Ao_lock()

        self.__tagfilter = []
        
        self.__addr_list = ("street_address", "city", "state", "postal_code", "country")
        self.__common    = {"title":"company_name", "phone":"phone_number", "email":"email_address"}
        
        self.old_title = appuifw.app.title
        appuifw.app.title = u"restobook"

        self.exit_flag = False
        appuifw.app.exit_key_handler = self.abort

        self.__options = [
            (u"Merge your restobook", self.merge),
            (u"Define a tag filter", self.add_tag_filter),
            (u"Add to your restobook", self.add_resto_unfiled),
            (u"Search your restobook", self.search_restos),
            (u"Set Y! Local location", self.assign_search_ylocal_location),
            ]

        self.__url_formats = {
            'yelp':'http://mobile.yelp.com/biz/%s',
            'ylocal':'http://local.yahoo.com/details?id=%s',
            'sfweekly':'http://www.sfweekly.com/Issues/%s/dining/eat_print.html'
        }

    #
    #
    #
    
    def define_url_formats (self, user_formats) :
        for key, fmt in user_formats.items() :
            self.__url_formats[key] = fmt

    #
    #
    #

    def loop (self) :

        if self.setup() :
            self.run()
            
        self.lock.wait()

    #
    #
    #
    
    def abort(self) :
        self.store_cfg()
        self.lock.signal()
        self.old_title = appuifw.app.title

    #
    #
    #

    def assign_search_ylocal_location (self) :

        loc = appuifw.query(u"Default location for Y! Local search", "text")

        if loc == None :
            self.log("Cancelled. Moving along.")
            return True

        self.__cfg['search_ylocal_loc'] = loc
        
        if loc == '' :
            self.log("Default location reset.")
        else :
            self.log("Default location set to '%s'" % loc)
            
        return True
    
    #
    #
    #
    
    def enable_search_ylocal (self) :
        self.__cfg['search_ylocal'] = 'ok'

        self.log("Y! Local search enabled")        
        self.toggle_ysearch_menu()        

    #
    #
    #
    
    def disable_search_ylocal (self) :
        self.__cfg['search_ylocal'] = ''

        self.log("Y! Local search disabled")
        self.toggle_ysearch_menu()

    #
    #
    #
    
    def assign_menu (self) :
        appuifw.app.menu = self.__options
        
    #
    #
    #
    
    def toggle_ysearch_menu (self, startup=False) :

        if startup :
            self.__options.append(())

        #
            
        idx = len(self.__options) - 1
    
        if self.__cfg['search_ylocal'] == '' :
            self.__options[idx] = (u"Enable Y! Local search", self.enable_search_ylocal)
        else :
            self.__options[idx] = (u"Disable Y! Local search", self.disable_search_ylocal)
            
        self.assign_menu()
        
    #
    #
    #
    
    def open_cfg (self) :
        self.log("Reading configs")

        if not os.path.exists(self.__cfgpath) :
            self.__cfg = anydbm.open(self.__cfgpath, "c")
            self.__cfg.close()

        self.__cfg = anydbm.open(self.__cfgpath, "w")

        defaults = ['search_ylocal', 'search_ylocal_loc']

        for prop in defaults :
            if not self.__cfg.has_key(prop) :
                self.__cfg[prop] = ''
            
        return True

    #
    #
    #
    
    def store_cfg (self) :
        self.__cfg.close()
            
    #
    #
    #
    
    def setup (self) :

        self.log("restobook (%s) says hello!" % __version__)
        self.log("---------------------")
        
        if not self.open_cfg() :
            return False

        if not self.open_contacts() :
            return False

        if appuifw.query(u"Run in offline mode?", "query") != None :
            self.__offline = True

        if self.__offline == False :
            if not self.setup_network() :
                return False

        self.toggle_ysearch_menu(True);
        self.assign_menu()
        return True

    #
    #
    #
    
    def setup_network (self) :
        if not self.test_connection() :
            return False
        
        if not self.get_credentials() :
            return False
        
        if not self.test_credentials() :

            if appuifw.query(u"Login failed. try again?", "query") :
                self.change_credentials()
                return self.run()
            else :
                return False

        self.__network = True
        return True
    
    #
    #
    #

    def open_contacts (self) :

        if e32.in_emulator() :
            return True

        try :
            self.__db = contacts.open()
        except Exception, e :
            self.log("Failed to open address book, %s" % e)
            return False

        return True

    #
    #
    #
    
    def run (self) :
        self.clear_log()
        
        self.log("---------------------")
        self.log("Restobook is now ready to merge your del.icio.us posts tagged with 'restobook' and your address book. To begin, select 'Merge your restobook' from the Options menu.\n\n")
        self.log("You may also define addition tag filters to limit the merge with. Select 'Define a tag filter' from the Options menu.\n\n")
        self.log("If you are out wandering about a spot a restaurant you'd like to add to your restobook, select 'Add to your restobook' from the Options menu. Restobook will add a private post to your del.icio.us links tagged with 'restobook:unfiled'.\n")
        self.log("---------------------")
        
    #
    #
    #

    def ensure_network (self) :

        if self.__offline == True :
            if not appuifw.query(u"Go online?", "query") :
                return False
            
            self.__offline = False
            
        if self.__network == True :
            return True

        if self.setup_network() :
            return True

        return False

    #
    #
    #
    
    def add_resto_unfiled (self) :

        if not self.ensure_network() :
            return False

        #
        # Hello world
        #
        
        resto = appuifw.query(u"Resto name", "text")

        if resto == None :
            self.log("Okay, well there's nothing to add then. Moving along!")
            return False

        post_data = {'url':'', 'description':'', 'extended':'', 'tags':'restobook:unfiled', 'shared':'no'}
        
        #
        # Check to see if this is a Y! local ID
        #
        
        local = re.match("y:(\d+)", resto)

        if local :
            local_id   = local.groups()[0]
            resto_data = self.retrieve_ylocal_id(local_id)

            if resto_data == None :
                self.error("Hrm, there was a problem retrieving restaurant data from Y! Local")
                return False

            self.merge_resto_post_data(post_data, resto_data)
            
        else :

            post_data['description'] = resto
            post_data['url']         = "http://del.icio.us/search/?all=%s" % resto

            #
            # Try to search for the restaurant in Y! local
            #
            
            if self.__cfg['search_ylocal'] != '' :

                if self.__cfg['search_ylocal_loc'] != '' :

                    self.log("Searching Y! Local for '%s' in %s" % (resto, self.__cfg['search_ylocal_loc']))
                         
                    try :
                        resto_data = self.search_ylocal(resto, None, None, self.__cfg['search_ylocal_loc'])
                        self.merge_resto_post_data(post_data, resto_data)
                        local = True
                         
                    except Exception, e:
                        self.error("Y! Local search failed, %s" % e)
                else :
                    self.error("Can't search Y! Local for '%s' because no default location has been set. You can do this from > Options > Set Y! Local location")

        #
        # Tags
        #
        
        user_tags = appuifw.query(u"Tags? (in addition to 'restobook:unfiled')", "query")

        if user_tags != None :
            tags = self.select_tags()
            sep  = u" "
            
            post_data['tags'] += " %s" % sep.join(tags)
            
        #
        # Add to local store
        #

        if local :

            if appuifw.query(u"Add to your address book now", "query") != None :
                if self.store_resto(resto_data) :
                    self.log("%s has been added to your address book" % post_data['description'])
                else :
                    self.error("Hrm, there was a problem updating your address book")
        
        #
        # Post!
        #

        # self.log("POST : %s" % post_data)
        
        xml = self.post_resto(post_data)

        if xml == None or xml.attrib['code'] != 'done' :
            self.error("Failed to post '%s' : %s" % (resto))
            return False

        self.log("Posted '%s' to del." % post_data['description'])

        #
        
        return True

    #
    #
    #
    
    def merge_resto_post_data (self, post_data, resto_data) :

        addr   = ":address \"%s %s %s\";" % (resto_data['street_address'], resto_data['city'], resto_data['state'])
        phone  = ":phone \"%s\";" % resto_data['phone_number']
            
        post_data['url']         = resto_data['uri']['homepage']
        post_data['description'] = resto_data['company_name']
        post_data['extended']    = "%s %s" % (addr, phone)
                  
        if resto_data['ylocal_id'] != '' :
            
            ylocal = " :ylocal \"%s\";" % resto_data['ylocal_id']            
            post_data['extended'] = "%s %s" % (post_data['extended'], ylocal)
        
    #
    #
    #
    
    def post_resto (self, post_data) :

        params = urllib.urlencode(post_data)
        req    = "/v1/posts/add?%s" % params

        xml = self.execute_delicious_request(req)
        return xml
    
    #
    #
    #
    
    def add_tag_filter (self) :
        
        for tag in self.select_tags() : 
            if not tag in self.__tagfilter :
                self.log("Added '%s' to tag filters" % tag)
                self.__tagfilter.append(tag)
            
        return True

    #
    #
    #
    
    def merge (self) :

        if not self.ensure_network() :
            return False

        self.define_merge_criteria()
        self.define_date_criteria()
        
        #
        
        dates = self.fetch_dates()

        if len(dates) == 0 :
            self.error("Hrm...there was a problem rounding up restobook dates")
            return False
            
        posts = self.fetch_posts(dates)

        if posts == None or len(posts) == 0 :
            self.error("No suitable posts were found.")
            return False
        
        restos = self.parse_posts(posts)

        if e32.in_emulator() :
            self.dump_restos(restos)
        else :
            for resto in restos :
                self.store_resto(resto)

        self.log("---------------------")
        self.log("All done! Mange! Mange!!")

        self.__tags = []
        return True

    #
    #
    #

    def define_merge_criteria (self) :
        opts   = [u"Merge (default)", u"Skip", u"Ask first"]
        prompt = u"How should restobook handle existing records"
        merge  = appuifw.popup_menu(opts, prompt)

        if merge == None :
            merge = 0
            
        self.__cfg["merge"] = unicode(merge)

    #
    #
    #

    def define_date_criteria (self) :
        response = appuifw.query(u"Only fetch most recent posts?", "query")
        most_recent = 1
        
        if response == None :
            most_recent = 0

        self.__cfg["most_recent"] = unicode(most_recent)
        
    #
    #
    #
    
    def dump_restos (self, restos) :
        for r in restos :
            self.dump_resto(r)

    #
    #
    #
    
    def dump_resto (self, r) :
            self.log("%s (dump) :" % r['company_name'])
            self.log("--------------------------------")
            
            for k, v in r.items() :
                if k == 'uri' :
                    for label, link in r[k].items() :
                        self.log("%s (%s)" % (link, label))
                else :
                    self.log("%s : %s" % (k, v))

            self.log("--------------------------------")
            
    #
    #
    #

    def change_credentials (self) :

        for prop in ("user", "pswd") :
            if self.__cfg.has_key(prop) :
                del(self.__cfg[prop])

        self.store_cfg()
        return self.get_credentials()

    #
    #
    #
    
    def get_credentials (self) :

        user = None
        pswd = None
        
        if not self.__cfg.has_key("user") :
            user = appuifw.query(u"your del. username", "text")

            if user == None :
                return False

            self.__cfg["user"] = user
        
        if not self.__cfg.has_key("pswd") :
            pswd = appuifw.query(u"your del. pswd", "text")

            if pswd == None :
                return False

            self.__cfg["pswd"] = pswd

        return True
            
    #
    #
    #

    def test_credentials (self) :
        self.log("Testing your del. login")
    
        res = self.execute_delicious_request("/v1/posts/update")

        if res == None :
            return False

        return True

    #
    #
    #
    
    def test_connection (self) :
        self.log("Testing your network connection")
        return True
    
        try :
            conn = httplib.HTTPConnection('del.icio.us')
            conn.request('GET', '/')
        except Exception, e :
            self.error("Can not find the network love (%s) :-(" % e)
            return False

        return True
    
    #
    #
    #
    
    def fetch_posts (self, dates) :
        posts = []
        
        for dt in dates :

            tf = "restobook"

            for t in self.__tagfilter :
                tf += " %s" % t
                
            #
            
            if dt == "recent" :
                self.log("Fetching recent posts tagged with %s" % tf)
                params = urllib.urlencode({'tag':'restobook'})
            else :
                self.log("Fetching posts tagged with %s on %s" % (tf, dt))
                params = urllib.urlencode({'tag':'restobook', 'dt':dt})

            req = "/v1/posts/get?%s" % params

            xml = self.execute_delicious_request(req)

            if xml == None :
                self.error("Hrm...couldn't get posts for %s" % dt)
                continue
            
            xpath = compile("/posts/post")
            
            for p in xpath.evaluate(xml) :

                if self.add_link(p) and self.test_tag_filters(p) :                    
                    posts.append(p)

        return posts
    
    #
    #
    #
    
    def fetch_dates (self) :

        if int(self.__cfg["most_recent"]) == 1 :
            return [u"recent"]

        #
        
        self.log("Fetching dates tagged with restobook")        
        dates = []

        req = "/v1/posts/dates?tag=restobook"
        xml = self.execute_delicious_request(req)

        if xml == None :
            return dates
        
        xpath = compile("dates/date")
        res   = xpath.evaluate(xml)
        
        for dt in res :
            dates.append(dt.attrib['date'])

        return dates
    
    #
    #
    #
    
    def execute_delicious_request (self, req) :

        base64string = base64.encodestring('%s:%s' % (self.__cfg['user'], self.__cfg['pswd']))[:-1]
        headers = {"Authorization":"Basic %s" % base64string}
        
        url = "api.del.icio.us"
    
        conn = httplib.HTTPSConnection(url)
        conn.request('GET', req, None, headers)

        res = conn.getresponse()

        if res.status != 200 :
            self.error(res.reason)
            return None

        content = res.read()
        # print content
        
        return XML(content)

    #
    #
    #
    
    def geocode_address (self, addr, resto) :
        self.log("Trying to geocode %s" % addr)
        
        params = urllib.urlencode({"appid":"restobook", "location":addr})
        req    = "/MapsService/V1/geocode?%s" % params

        xml = self.execute_yahoo_request(req)
        res = xml.find("{urn:yahoo:maps}Result")

        if res == None :
            return False

        findme = list(self.__addr_list)
        findme.append("latitude")
        findme.append("longitude")

        for prop in findme :

            ykey = prop.title()
            
            if prop == "street_address" :
                ykey = "Address"
            elif prop == "postal_code" :
                ykey = "Zip"
            else :
                pass

            el   = "{urn:yahoo:maps}%s" % ykey
            node = res.find(el)

            if node == None:
                resto[prop.lower()] = ""                                
            else :
                resto[prop.lower()] = node.text.title()

        #
        # see if the raw address has a cross street
        #
        
        if resto.has_key('city') :
            pattern = "@\s+?(.*)\s+%s" % resto['city']
            match   = re.search(pattern, addr)

            if match :
                resto['cross street'] = match.groups()[0]

        # 
        #
        #
        
        resto['geo_address'] = addr
        return True

    #
    #
    #

    def retrieve_ylocal_id (self, id) :

        xmlns = "urn:yahoo:lcl"
        
        req = "/LocalSearchService/V3/localSearch?appid=restobook&listing_id=%s&query=*" % id
        xml = self.execute_yahoo_request(req)

        xp  = "{%s}Result" % xmlns
        
        try :
            res = xml.find(xp)
        except Exception, e :
            self.error("Failed to parse response, %s" % e)
            return None

        if res == None :
            return None

        return self.parse_ylocal_result(res)

    #
    #
    #
    
    def parse_ylocal_result (self, res) :

        xmlns = "urn:yahoo:lcl"
        
        local = {"Title":"company_name", "Address":"street_address", "City":"", "State":"", "Phone":"phone_number", "Latitude":"", "Longitude":"", "BusinessUrl":""}
        resto = {'uri':{}, 'ylocal_id':''}

        #
        
        if res.attrib.has_key('id') :
            resto['ylocal_id'] = res.attrib['id']
        else :
            url     = res.find("{%s}Url" % xmlns)
            matches = re.findall("\?id=(\d+)", url.text)

            if matches :
                resto['ylocal_id'] = matches[0]

        #
        
        for prop in local.keys() :

            el   = "{%s}%s" % (xmlns, prop)
            node = res.find(el)
            
            if node == None :
                continue

            if prop == 'BusinessUrl' :
                if node.text == None :
                    title = res.find("{%s}Title" % xmlns)
                    resto['uri']['homepage'] = "http://del.icio.us/search?all=%s" % title.text
                else :
                    resto['uri']['homepage'] = node.text                                        

            else :
                
                if local[prop] != "" :
                    resto_key = local[prop]
                else :
                    resto_key = prop.lower()

                resto[resto_key] = node.text.title()

        #
        
        if resto.has_key("latitude") and resto.has_key("longitude") :
            resto['geoposition'] = "%s, %s" % (resto['latitude'], resto['longitude'])
            del(resto['latitude'])
            del(resto['longitude'])
            
        return resto
    
    #
    #
    #
    
    def execute_yahoo_request(self, req) :

        url = "api.local.yahoo.com"
        
        conn = httplib.HTTPConnection(url)
        conn.request('GET', req)

        res = conn.getresponse()

        if res.status != 200 :
            self.error(res.reason)
            return None

        content = res.read()
        # print content

        try :
            return XML(content)
        except Exception, e :
            self.error("Failed to parse response for %s, %e" % (req, e))
            return None
        
    #
    #
    #
    
    def parse_posts (self, posts) :
        self.log("Parsing %d results" % len(posts))

        parsed = []

        known_url_labels = self.__url_formats.keys()
        
        for r in posts :

            title = r.attrib["description"]
            self.log("Parsing data for %s" % title)
            
            resto = self.parse_n3card(r.attrib["extended"])

            #

            # sh : short(hand)
            # lh : longhand
            
            for sh, lh in self.__common.items() :
                if not resto.has_key(sh) :
                    continue

                if not resto.has_key(lh) :
                    resto[lh] = resto[sh]

                del(resto[sh])

            #

            link = r.attrib["href"]
            
            resto['company_name'] = r.attrib["description"]
            resto['tags']  = r.attrib["tag"]
            resto['uri']   = {"homepage":link}

            for key in resto.keys() :

                if key in known_url_labels :
                    fmt = self.__url_formats[key]
                    resto['uri'][key] = fmt % resto[key]
                    del(resto[key])
                    
            if resto.has_key("latitude") and resto.has_key("longitude") :
                coords = "%s, %s" % (resto['latitude'], resto['longitude'])
                params = urllib.urlencode({'addr':resto['geo_address']})

                resto['uri']['map'] = "http://maps.yahoo.com/py/maps.py?%s" % params
                resto['geoposition'] = coords
                
                del(resto['latitude'])
                del(resto['longitude'])
                del(resto['geo_address'])                                
                
            hex = md5.md5(link).hexdigest()
            resto['uri']['del.icio.us'] = "http://del.icio.us/url/%s" % hex
            
            parsed.append(resto)

        return parsed

    #
    #
    #

    def add_link (self, post) :
        merge = int(self.__cfg["merge"])
        
        if not merge :
            return True
        
        link  = post.attrib["href"]
        title = post.attrib["description"]
        
        if not self.find_resto(link) :
            return True

        if merge == 1 :
            self.log("Skip %s" % title)
            return False

        if not appuifw.query(u"Merge %s?" % title, "query") :
            self.log("Skip %s" % title)
            return False

        return True

    #
    #
    #

    def test_tag_filters (self, post) :

        # There are no tag filters, so yes
        
        if len(self.__tagfilter) == 0 :
            return True
        
        tags = post.attrib["tag"]

        for tag in tags.split(" ") :
            if tag in self.__tagfilter :
                return True

        return False
            
    #
    #
    #
    
    def store_resto (self, data) :

        if e32.in_emulator() :
            self.dump_resto(data)
            return True
        
        resto = self.find_resto(data['uri']['homepage'])
        
        if resto == None :
            self.log("Adding %s" % data['company_name'])
            self.add_resto(data)
        else :
            self.log("Updating %s" % data['company_name'])   
            self.update_resto(resto, data)

        return True
    
    #
    #
    #
    
    def find_resto (self, uri) :

        if e32.in_emulator() :
            return None
        
        res = self.__db.find(uri)

        if len(res) == 0 :
            return None

        return res[0]

    #
    #
    #

    def add_resto (self, data) :

        resto = self.__db.add_contact()
        basic = self.basic_fields()
                
        for prop in data :
            if prop in basic :
                resto.add_field(prop, value=data[prop])
            elif prop == 'reservations' :
                resto.add_field('phone_number', value=data[prop], label=prop.title())
            elif prop == 'uri' :
                for link in data['uri'] :
                    resto.add_field('url', value=data['uri'][link], label=link)
            else :
                resto.add_field('note', value=data[prop], label=prop)

        resto.commit()
        
    #
    #
    #

    def basic_fields (self) :
        fixme = []
        for k in self.__addr_list :
            fixme.append(k)
        for k in self.__common.values() :
            fixme.append(k)

        return fixme

    #
    #
    #

    def reserved_fields (self) :
        reserved = self.basic_fields()
        return reserved
    
    #
    #
    #
    
    def update_resto (self, resto, data) :
        resto.begin()

        basic = self.basic_fields()

        for prop in data :
            if prop in basic :
                found = resto.find(type=prop)

                if len(found) :
                    found[0].value = data[prop]
                else :
                    resto.add_field(prop, value=data[prop])

            # fix me :
            
            elif prop == 'reservations' :
                found = False

                for ph in resto.find(type='phone_number') :
                    if ph.label == prop.title() :
                        ph.value = data[prop]
                        found = True
                        break

                if not found :
                    resto.add_field('phone_number', value=data[prop], label=prop.title())
                
            elif prop == 'uri' :
                for link in data['uri'] :

                    found = 0
                    
                    for field in resto.find(type='url') :
                        if field.label == link :
                            field.value = data['uri'][link]
                            found = 1
                            break
                        
                    if not found :
                        resto.add_field('url', value=data['uri'][link], label=link)
            else :

                found = 0

                for field in resto.find(type='note') :
                    if field.label == prop :
                        field.value = data[prop]
                        found = 1
                        break

                if not found :
                    resto.add_field('note', value=data[prop], label=prop)

        resto.commit()
            
    #
    #
    #
    
    def parse_n3card (self, n3) :

        reserved = self.reserved_fields()
        resto    = {}

        if not n3.endswith(";") :
            n3 += ";"

        matches = re.findall('([^:]+)?:(?:([^\s]+)\s+"([^"]+)";\s?)', n3)

        for pair in matches :
            ns        = pair[0]
            predicate = pair[1]
            object    = pair[2]
            
            #

            if predicate == 'address' :
                if self.geocode_address(unicode(object), resto) :
                    continue
            if not ns == '' :

                if ns in reserved :
                    self.error("%s is a reserved label, skipping" % ns)
                    continue

                if not resto.has_key('__ns') :
                    resto['__ns'] = {}
                    
                if not resto['__ns'].has_key(ns) :
                    resto['__ns'][ns] = {}
                    
                resto['__ns'][ns][predicate] = object
            else :
                resto[predicate] = object

        #
        # Check to see whether we have no address data
        # but do have a Y! local ID - if so, try to fetch
        # address data from there
        #

        if not resto.has_key('street_address') and resto.has_key('ylocal') :

            yresto = self.retrieve_ylocal_data(resto[ylocal])

            if yresto != None :
                for key, value in yresto.items() :
                    if not resto.has_key(key) :
                        resto[key] = value

        #
        # Squish namespaced elements in to note format
        # If there's actually a way to add text with new
        # lines to a 'note' object, I can't figure out
        # how...
        #
        
        if resto.has_key('__ns') :

            for ns, data in resto['__ns'].items() :
                pairs = []
                sep = "; "
                
                for p, o in resto['__ns'][ns].items() :
                    pairs.append("%s \"%s\"" % (p, o))

                ns_key = "# %s" % ns
                resto[ns_key] = unicode(sep.join(pairs))
                
            del(resto['__ns'])

        #

        return resto

    #
    #
    #

    def search_restos (self) :

        if e32.in_emulator() :
            self.log("Search disabled in emulator")
            return True

        #
        
        self.__search = []
                        
        restos = self.__db.find('restobook')

        matching = self.select_tags()
        found    = []
        
        #
                    
        for resto in restos :
            matches = 0
            
            for note in resto.find(type="note") :
                if note.label=="tags" :
                    tags = note.value.split(" ")
                    
                    for t in matching :
                        if t in tags :
                            matches += 1
                    break
                
            if matches == len(matching) :
                found.append(resto)

        #

        if len(found) == 0 :
            if appuifw.query(u"No matches. Try another search?", "query") :
                return self.search_restos()

            return False            

        self.__search = found
        return self.display_search_results()

    #
    #
    #
    
    def display_search_results (self) :    
        
        labels = []

        for resto in self.__search :
            labels.append(resto.title)

        idx = appuifw.selection_list(labels)

        if idx == None :
            if appuifw.query(u"Try another search?", "query") :
                return self.search_restos()

            return False

        #

        return self.display_resto(self.__search[idx])    
        
    #
    #
    #
    
    def display_resto (self, resto) :

        path_vcard = u"c:\\restobook.vcard"

        try :
            fh = open(path_vcard, "w")
            fh.write(resto.as_vcard())
            fh.close()
        except Exception, e :
            self.error("Can't read %s (%s)" % (resto.title, e))
            return False

        try :
            content_handler = appuifw.Content_handler(self.display_search_results)
            content_handler.open(path_vcard)
        except Exception, e :
            self.error("Can't display %s (%s)" % (resto.title, e))
            return False
        
    #
    #
    #

    def select_tags (self, freetext=True, force_reload=False) :

        new_tag = u'- add a new tag'
        
        selected = []        
        tags     = self.collect_tags()

        if freetext and not new_tag in tags :
            tags.insert(0, new_tag)

        #
        
        idx = None
        
        if e32.in_emulator() :
            idx = appuifw.selection_list(tags)            

            # sigh...hacks to account for missing
            # emulation-ness in the emulator
            
            if idx != None :
                idx = [idx]
        else :
            idx = appuifw.multi_selection_list(tags, style="checkbox")

        #
        
        if idx != None :                
            for i in idx :

                if tags[i] == new_tag :
                    user_tags = appuifw.query(u"Enter your new tags separated by a space", "text")

                    for t in user_tags.split(' ') :
                        if not t in selected :
                            selected.append(t)
                            
                else :
                    selected.append(tags[i])        

        selected.sort()
        return selected
    
    #
    #
    #
    
    def collect_tags (self, force_reload=False) :

        list = []

        if e32.in_emulator() :
            return list

        if self.__tags != None and len(self.__tags) > 0 and not force_reload :
            return self.__tags
                
        for resto in self.__db.find('restobook') :

            for note in resto.find(type="note") :
                if note.label=="tags" :
                    tags = note.value.split(" ")

                    for t in tags :
                        if not t in list :
                            list.append(t)

        list.sort()

        self.__tags = list
        return list

    #
    #
    #

    def search_ylocal (self, term, lat=None, lon=None, loc=None) :

        params={'appid':'restobook','query':term}

        if loc != None :
            params['location'] = loc
        else :
            params['latitude']  = lat
            params['longitude'] = long
            params['radius']    = 2
            
        req = "/LocalSearchService/V2/localSearch?%s" % urllib.urlencode(params)
        xml = self.execute_yahoo_request(req)

        xmlns = "urn:yahoo:lcl"
        xp  = "{%s}Result" % xmlns

        places = xml.findall(xp)

        if places == None :
            self.log("Unable to find any matches for '%s'" % term)
            return None
        
        labels = []

        if len(labels) == 0 :
            self.log("Unable to find any matches for '%s'" % term)
            return None

        for r in places :
            t = r.find("{%s}Title" % xmlns)
            a = r.find("{%s}Address" % xmlns)            
            labels.append(unicode("%s, %s" % (t.text, a.text)))
            
        idx = appuifw.selection_list(labels)

        if idx == None :
            return None

        return self.parse_ylocal_result(places[idx])

    #
    #
    #

    def error (self, msg) :
        self.log(msg, 1)

    #
    #
    #

    def clear_log (self) :
        self.__msg = u""

    #
    #
    #
    
    def log (self, msg, error=0) :

        self.__msg += "%s\n" % msg
        self.write(unicode(self.__msg))

    #
    #
    #
    
    def write(self, txt) :

        if e32.in_emulator() :
            t = appuifw.Text()
            t.write(txt)
            appuifw.app.body = t

        else :
            appuifw.app.body = appuifw.Text(txt)
            
    #
    #
    #
    
if __name__ == "__main__" :
    app = Resto()
    app.loop()
